'''
Program name :  TurtleCurvesPMP
Author       :  Philip Murray-Pearce
Created      :  19/02/2015
Copyright    :  (c) Philip Murray-Pearce 2015
Licence      :  OSI Simplified BSD
'''


import turtle
import math as m
import tkinter.messagebox as mb
from fractions import Fraction


def trochoid(t, r1=200, r2=80, offset=20, eh = 'H', pen=['gray'], fill='white', \
    filled=False, centred=True, lr='R'):
    '''
    r1 is the (integer) radius of the outer circle.
    r2 is the (integer) radius of the inner circle.
    offset is the distance of the 'pen' from the centre of the inner circle.
      offset must not be the same as rr (try 0.01 bigger or smaller) for a pattern to draw properly.
    The number of complete 360 turns the inner circle must make to create a pattern with start and
      finish correctly joined is the numerator of r2/r1 (reduced as far as possible) and the number
      of revolutions it makes within in the outer circle to do so is the denominator.
      Not all combinations of r1, r2, and offset produce a complete pattern.
    eh is either H (hypotrochoid) or E (epitrochoid).
    pen is a list of one or more colours which are repeated during drawing.
    '''
    if lr not in ['L','R']:
        mb.showerror('Trochoid parameters error!', 'Direction (lr) must be R or L.')
        return
    if eh not in ['H','E']:
        mb.showerror('Trochoid parameters error!', \
            'Type (eh) must be H (hypotrochoid) or E (epitrochoid).')
        return
    if r1 == r2:
        mb.showerror('Trochoid parameters error!', 'The two circles must not be the same size.')
        return
    if r1 <= 0 or r2 <= 0:
        mb.showerror('Trochoid parameters error!', \
            'The radii of both circles must be greater than 0.')
        return
    h = 1
    if eh == 'E':
        h = -1
    d = 1
    if lr == 'L':
        d = -1
    k = Fraction(r2, r1).numerator
    a = r1 - r2*h + offset
    if r2 > r1:
        a = r2*h - r1 - offset
    x0 = r1 - r2*h + offset
    y0 = 0
    x1 = (r1-r2*h)*m.cos(m.radians(1)) + offset*m.cos((r1-r2*h)/r2*m.radians(1))
    y1 = (r1-r2*h)*m.sin(m.radians(1)) - offset*m.sin((r1-r2*h)/r2*m.radians(1))
    sd1 = m.sqrt((x1-x0)**2 + (y1-y0)**2)
    firstturnangle = m.degrees(m.atan((y1-y0)/(x1-x0)))
    t.pu()
    if centred:
        t.lt(90*d)
        t.fd(a)
        t.rt(90*d)
    t.rt((90 - m.fabs(firstturnangle)) * d)
    t.pd()
    if filled:
        t.fillcolor(fill)
        t.begin_fill()
    for i in range(0, 360 * k + 1):
        x2 = (r1-r2*h)*m.cos(m.radians(i+2)) + offset*m.cos((r1-r2*h)/r2*m.radians(i+2))
        y2 = (r1-r2*h)*m.sin(m.radians(i+2)) - offset*m.sin((r1-r2*h)/r2*m.radians(i+2))
        sd2 = m.sqrt((x2-x1)**2 + (y2-y1)**2)
        chord = m.sqrt((x2-x0)**2 + (y2-y0)**2)
        turnangle = 180 - m.degrees(m.acos((sd1*sd1 + sd2*sd2 - chord*chord)/(2*sd1*sd2)))
        if i == 360 * k:
            finalturnangle = turnangle
            finalsd1 = sd1
        t.pd()
        if i == 0:
            t.pu()
        t.pencolor(pen[i % len(pen)])
        t.fd(sd1)
        t.rt(turnangle * d)
        x0 = x1
        y0 = y1
        x1 = x2
        y1 = y2
        sd1 = sd2
        if i >= 360 * k + 1:
            t.lt(turnangle * d)
            t.bk(sd1)
    if filled:
        t.end_fill()
    t.pu()
    t.lt(finalturnangle * d)
    t.bk(finalsd1)
    t.lt((90 - m.fabs(firstturnangle)) * d)
    t.pu()
    if centred:
        t.rt(90*d)
        t.fd(a)
        t.lt(90*d)


def rhodonea(t, size=200, numerator=5, denominator=1, pen=['gray'], fill='white', filled=False, \
    centred=True, lr='R'):
    '''
    Rhodonea is the original name for rose curve.
    size is the radius of the circle containing the rose curve.
    numerator and denominator are any integers.
      If the denominator is 1, the shape drawn is a 'flower' with 'petals'.
      If the numerator is greater than the denominator, this gives broader 'petals'.
      If the numerator is less than the denominator, this gives circle-based 'rose-window' shapes.
    pen is a list of one or more colours which are repeated during drawing.
    '''
    if lr not in ['L','R']:
        mb.showerror('Rose curve parameters error!', 'Direction (lr) must be R or L.')
        return
    if size <= 0 or numerator <= 0 or denominator <= 0:
        mb.showerror('Rose curve parameters error!', \
            'The values of size, numerator, and denominator must all be greater than 0.')
        return
    if not (numerator == int(numerator) and denominator == int(denominator)):
        mb.showerror('Rose curve parameters error!', \
            'The values of numerator and denominator must both be integers.')
        return
    d = 1
    if lr == 'L':
        d = -1
    p = numerator / denominator
    k = denominator
    if numerator % 2 == 1 and denominator % 2 == 1:
        k = k / 2
    a = size
    x0 = size
    y0 = 0
    x1 = m.cos(m.radians(p)) * m.cos(m.radians(1)) * size
    y1 = m.cos(m.radians(p)) * m.sin(m.radians(1)) * size
    sd1 = m.sqrt((x1-x0)**2 + (y1-y0)**2)
    firstturnangle = m.degrees(m.atan((y1-y0)/(x1-x0)))
    t.pu()
    if centred:
        t.lt(90*d)
        t.fd(a)
        t.rt(90*d)
    t.rt((90 - m.fabs(firstturnangle)) * d)
    t.pd()
    if filled:
        t.fillcolor(fill)
        t.begin_fill()
    for i in range(int(360 * k)):
        x2 = m.cos(m.radians(p*(i+2))) * m.cos(m.radians(i+2)) * size
        y2 = m.cos(m.radians(p*(i+2))) * m.sin(m.radians(i+2)) * size
        sd2 = m.sqrt((x2-x1)**2 + (y2-y1)**2)
        chord = m.sqrt((x2-x0)**2 + (y2-y0)**2)
        turnangle = 180 - m.degrees(m.acos((sd1*sd1 + sd2*sd2 - chord*chord)/(2*sd1*sd2)))
        t.pd()
        t.pencolor(pen[i % len(pen)])
        t.fd(sd1)
        t.rt(turnangle * d)
        x0 = x1
        y0 = y1
        x1 = x2
        y1 = y2
        sd1 = sd2
    if filled:
        t.end_fill()
    t.lt((90 - m.fabs(firstturnangle)) * d)
    t.pu()
    if centred:
        t.rt(90*d)
        t.fd(a)
        t.lt(90*d)


def lissajous(t, scale=100, width=2, height=1, angfreq1=5, angfreq2=4, phasediff=0, \
    pen=['gray'], lr='R'):
    '''
    A Lissajous curve describes complex harmonic motion.
    scale is the relative size of the Lissajous curve.
    width and height are any integers.
      width / height gives the ratio of width to height in the drawn shape.
    angfreq1 and angfreq2 are any integers.
      angfreq1 gives the number of horizontal 'lobes' in the drawn shape.
      angfreq2 gives the number of vertical 'lobes' in the drawn shape.
      If both angfreq1 and angfreq2 are odd, the curve does not complete.
      If angfreq2 = 1, the curve does not complete accurately (? due to inaccuracy in sin function).
    phasediff should be left as 0 to produce a complete curve.
    pen is a list of one or more colours which are repeated during drawing.
    '''
    af1 = Fraction(angfreq1, angfreq2).numerator
    af2 = Fraction(angfreq1, angfreq2).denominator
    angfreq1 = af1
    angfreq2 = af2
    if lr not in ['L','R']:
        mb.showerror('Lissajous curve parameters error!', 'Direction (lr) must be R or L.')
        return
    if scale <= 0 or width <=0 or height <= 0 or angfreq1 <= 0 or angfreq2 <= 0:
        mb.showerror('Lissajous curve parameters error!', \
            'The values of scale, width, height, angfreq1, and angfreq2 must all be greater than 0.')
        return
    if not (width == int(width) and height == int(height)):
        mb.showerror('Lissajous curve parameters error!', \
            'The values of width and height must both be integers.')
        return
    if not (angfreq1 == int(angfreq1) and angfreq2 == int(angfreq2)):
        mb.showerror('Lissajous curve parameters error!', \
            'The values of angfreq1 and angfreq2 must both be integers.')
        return
    if angfreq1 % 2 == 1 and angfreq2 % 2 == 1:
        mb.showerror('Lissajous curve parameters error!', \
            'At least one of angfreq1 and angfreq2 must be an even number.')
        return
    if angfreq2 >= angfreq1:
        mb.showerror('Lissajous curve parameters error!', \
            'The value of angfreq1 must be greater than that of angfreq2.')
        return
    if not (angfreq2 == 1 or angfreq1-angfreq2 == 1):
        mb.showwarning('Lissajous curve parameters error!', \
            'The value of angfreq2-angfreq1 or of angfreq2 must be 1 in this version.')
        return
    d = 1
    if lr == 'L':
        d = -1
    kk = 1
    if angfreq2 == 1 and 180%angfreq1 != 0:
        while 180*kk%angfreq1 != 0:
            kk += 1
    elif angfreq1-angfreq2 == 1:
        kk = 1 + angfreq1//8
    x0 = width * m.sin(m.radians(phasediff)) * scale
    y0 = 0
    x1 = width * m.sin(m.radians(angfreq1*1/kk + phasediff)) * scale
    y1 = height * m.sin(m.radians(angfreq2*1/kk)) * scale
    sd1 = m.sqrt((x1-x0)**2 + (y1-y0)**2)
    firstturnangle = m.degrees(m.atan((y1-y0)/(x1-x0)))
    d1 = d
    t.rt((90 + firstturnangle) * d1)
    t.pd()
    for i in range(360*kk+1):
        x2 = width * m.sin(m.radians(angfreq1*(i+2)/kk + phasediff)) * scale
        y2 = height * m.sin(m.radians(angfreq2*(i+2)/kk)) * scale
        sd2 = m.sqrt((x2-x1)**2 + (y2-y1)**2)
        chord = m.sqrt((x2-x0)**2 + (y2-y0)**2)
        zz = (sd1*sd1 + sd2*sd2 - chord*chord)/(2*sd1*sd2)
        if zz > 1:  # Applies a correction to zz if it is outside the range -1 >= zz <= 1.
            zz = 1
        elif zz < -1:
            zz = -1
        if angfreq2 == 1:
            if i%((180*kk)//angfreq1) == 0 \
               and not (i == 0 or i == 90*kk or i == 270*kk or i == 360*kk):
                d = -d
        elif angfreq1 - angfreq2 == 1:
            if i == 180*kk or i == 360*kk:
                d = -d
        turnangle = 180 - m.degrees(m.acos(zz))
        if i == 360*kk:
            finalturnangle = turnangle
            finalsd1 = sd1
        t.pd()
        t.pencolor(pen[i % len(pen)])
        t.fd(sd1)
        t.rt(turnangle * d)
        x0 = x1
        y0 = y1
        x1 = x2
        y1 = y2
        sd1 = sd2
    t.pu()
    t.lt(finalturnangle * d)
    t.bk(finalsd1)
    t.lt((90 + firstturnangle) * d1)


def archimedeanspiral(t, turns=10, gap=1, step=10, pen=['black'], lr='R'):
    '''
    An Archimedean Spiral is a spiral with a constant gap between successive turns.
    turns is the number of turns (or complete rotations) of the spiral.
    gap is the constant distance between successive turns of the spiral.
    step determines the accuracy of the drawing of the spiral, higher values being less accurate.
    '''
    if lr not in ['L','R']:
        mb.showerror('Spiral parameters error!', 'Direction (lr) must be R or L.')
        return
    if turns <= 0 or gap <= 0 or step <= 0:
        mb.showerror('Spiral parameters error!', \
            'Turns, gap, and step must all be greater than 0.')
        return
    d = 1
    if lr == 'L':
        d = -1
    x0 = 0
    y0 = 0
    x1 = m.cos(m.radians(1)) * gap/10
    y1 = m.sin(m.radians(1)) * gap/10
    sd1 = m.sqrt((x1-x0)**2 + (y1-y0)**2)
    for i in range(0, 360 * turns, step):
        x2 = i * m.cos(m.radians(i)) * gap/10
        y2 = i * m.sin(m.radians(i)) * gap/10
        sd2 = m.sqrt((x2-x1)**2 + (y2-y1)**2)
        chord = m.sqrt((x2-x0)**2 + (y2-y0)**2)
        turnangle = 180 - m.degrees(m.acos((sd1*sd1 + sd2*sd2 - chord*chord)/(2*sd1*sd2)))
        t.pd()
        t.pencolor(pen[int(i/step) % len(pen)])
        t.fd(sd1)
        t.rt(turnangle * d)
        x0 = x1
        y0 = y1
        x1 = x2
        y1 = y2
        sd1 = sd2


def logarithmicspiral(t, turns=10, gap=1, step=10, pen=['black'], lr='R'):
    '''
    A Logarithmic Spiral is a spiral with an increasing gap between successive turns.
    turns is the number of turns (or complete rotations) of the spiral.
    gap is a constant controlling the 'tightness' of the spiral, higher values being 'tighter'.
    step determines the accuracy of the drawing of the spiral, higher values being less accurate.
    '''
    if lr not in ['L','R']:
        mb.showerror('Spiral parameters error!', 'Direction (lr) must be R or L.')
        return
    if turns <= 0 or gap <= 0 or step <= 0:
        mb.showerror('Spiral parameters error!', \
            'Turns, gap, and step must all be greater than 0.')
        return
    d = 1
    if lr == 'L':
        d = -1
    gap = gap*360
    x0 = 0
    y0 = 0
    x1 = m.exp(1/gap) * m.cos(m.radians(1))
    y1 = m.exp(1/gap) * m.sin(m.radians(1))
    sd1 = m.sqrt((x1-x0)**2 + (y1-y0)**2)
    for i in range(0, 360 * turns, step):
        x2 = m.exp(i/gap) * m.cos(m.radians(i))
        y2 = m.exp(i/gap) * m.sin(m.radians(i))
        sd2 = m.sqrt((x2-x1)**2 + (y2-y1)**2)
        chord = m.sqrt((x2-x0)**2 + (y2-y0)**2)
        turnangle = 180 - m.degrees(m.acos((sd1*sd1 + sd2*sd2 - chord*chord)/(2*sd1*sd2)))
        t.pd()
        t.pencolor(pen[int(i/step) % len(pen)])
        t.fd(sd1)
        t.rt(turnangle * d)
        x0 = x1
        y0 = y1
        x1 = x2
        y1 = y2
        sd1 = sd2


'''

Main Python Turtle Commands

t.clear()       Clear screen
t.home()        Put turtle at screen centre pointing up
t.speed()       Set speed of turtle movement - has one parameter (0-10)
t.st()          Show turtle
t.ht()          Hide turtle (drawing is faster with turtle hidden)
t.setpos()      Set turtle position - has 2 parameters for x and y position
t.seth()        Set turtle heading - has 1 parameter (0-360)
t.fd()          Move forwards - has 1 parameter for distance
t.bk()          Move backward - has 1 parameter for distance
t.rt()          Turn right - has 1 parameter for number of degrees
t.lt()          Turn left  - has 1 parameter for number of degrees
t.circle()      Draw an arc - has 2 main parameters for radius and angle turned
t.pd()          Pen down
t.pu()          Pen up
t.pensize()     Pen width   - has 1 parameter for width
t.pencolor()    Pen colour  - has 1 constant or 3 RGB values as parameters
t.fillcolor()   Fill colour - has 1 constant or 3 RGB values as parameters
t.begin_fill()  Shapes drawn between this and t.end_fill are filled in
t.end_fill()    Ends filling in started by t.begin_fill
t.dot()         Put a dot at the current position
t.stamp()       Put a turtle at the current position and heading
t.write()       Writes text - details of text, alignment, font, etc. as parameters


PMP Turtle Curves Commands

trochoid(t, r1=200, r2=80, offset=20, eh = 'H', pen=['gray'], fill='white',
    filled=False, centred=True, lr='R')
rhodonea(t, size=200, numerator=5, denominator=1, pen=['gray'], fill='white', filled=False, \
    centred=True, lr='R')
archimedeanspiral(t, turns=10, gap=1, step=10, pen=['black'], lr='R')
logarithmicspiral(t, turns=10, gap=1, step=10, pen=['black'], lr='R')

'''
